from django.http import HttpRequest

from pydantic import BaseModel

from .base import Control
from newsflash.callback import Callback, construct_callback


class ButtonContext(BaseModel):
    id: str
    text: str
    callback: Callback | None


class Button(Control):
    template_name: str = "control/button"
    text: str

    def _build(self, request: HttpRequest) -> ButtonContext:
        assert self.id is not None

        if "on_click" in self.__class__.__dict__:
            callback = construct_callback(self.__class__.on_click, "click", "click")
        else:
            callback = None

        return ButtonContext(
            id=self.id,
            text=self.text,
            callback=callback,
        )

    def on_click(self, *args, **kwargs) -> None:
        ...



import os
from typing import Type, TypeVar
from newsflash.widgets.layout.base import Layout
from newsflash.widgets.base import Widget
from newsflash.widgets import Notifications, Columns

from django.http import HttpRequest


T = TypeVar("T")


class App:
    layout: Layout

    def __init__(self) -> None:
        self.layout = self.compose()
        self.layout.main_element = True

    def compose(self) -> Layout:
        return Columns(children=[])

    def query_one(self, id: str, type: Type[T]) -> T | None:
        notifications = Notifications()
        if id == "notifications" and isinstance(notifications, type):
            return notifications

        return self.layout.query_one(id, type)

    def query_type(self, id: str) -> Type[Widget] | None:
        return self.layout.query_type(id)

    def render(self, request: HttpRequest) -> str:
        print(type(self.layout))
        return self.layout.render(request)

    @classmethod
    def run(cls):
        os.environ.setdefault("DJANGO_SETTINGS_MODULE", "web.settings")
        try:
            from django.core.management import execute_from_command_line
        except ImportError as exc:
            raise ImportError(
                "Couldn't import Django. Are you sure it's installed and "
                "available on your PYTHONPATH environment variable? Did you "
                "forget to activate a virtual environment?"
            ) from exc

        from web.app.urls import set_urlpatterns

        set_urlpatterns(cls)

        execute_from_command_line(["manage.py", "runserver"])



from pydantic import BaseModel


class Padding(BaseModel):
    ps: float
    pt: float
    pe: float
    pb: float


class Point(BaseModel):
    x: float
    y: float


def chart_space_to_xy(
    max_width: float,
    max_height: float,
    padding: Padding,
    width_in_px: float,
    height_in_px: float,
    point: Point,
) -> Point:
    x_frac = point.x / max_width
    y_frac = 1 - (point.y / max_height)

    chart_width_px = width_in_px - padding.ps - padding.pe
    chart_height_px = height_in_px - padding.pt - padding.pb

    x_px = padding.ps + (x_frac * chart_width_px)
    y_px = padding.pt + (y_frac * chart_height_px)

    return Point(x=x_px, y=y_px)


def calculate_ticks(x_end: float, x_start: float) -> tuple[float, float]:
    tick = 1

    x_range = x_end - x_start
    while x_range / tick > 8:
        if x_range / tick > 50:
            tick *= 10
        elif x_range / tick > 15:
            tick *= 5
        elif x_range / tick > 8:
            tick *= 2

    return tick, tick / 5



from typing import Callable, get_args, ForwardRef
from inspect import signature, getsourcefile
from runpy import run_path

from .widgets.control.base import Control
from .widgets.chart.type import Chart

from pydantic import BaseModel


class Callback(BaseModel):
    endpoint_name: str
    trigger_event: str
    inputs: list[str]
    targets: list[str]
    target_wrapper_ids: str


def construct_callback(
    callback_fn: Callable,
    endpoint_name: str,
    trigger_event: str,
) -> Callback:
    sig = signature(callback_fn)
    input_ids: list[str] = []
    target_ids: list[str] = []

    for param in sig.parameters:
        if param == "self" or param == "selected":
            continue
        annotation = sig.parameters[param].annotation
        args = get_args(annotation)

        try:
            widget_type = args[0]
            widget_id = args[1]
        except IndexError:
            continue

        if isinstance(widget_type, ForwardRef):
            source_file = getsourcefile(callback_fn)
            assert source_file is not None
            file_globals = run_path(source_file)
            widget_type = file_globals[widget_type.__forward_arg__]

        if issubclass(widget_type, Control):
            input_ids.append(widget_id + "-input")

        if issubclass(widget_type, Chart):
            target_ids.append(widget_id)

    return Callback(
        endpoint_name=endpoint_name,
        trigger_event=trigger_event,
        inputs=input_ids,
        targets=target_ids,
        target_wrapper_ids=", ".join([f"#{target}-wrapper" for target in target_ids]),
    )



from typing import TypeVar

from django.http import HttpRequest
from django.template.loader import render_to_string
from pydantic import BaseModel


T = TypeVar("T", bound="Widget")


class Widget(BaseModel):
    template_name: str = ""
    id: str | None = None
    width_in_px: float = 0.0
    height_in_px: float = 0.0
    span: int = 1
    swap_oob: bool = False

    class Config:
        validate_assignment = True

    class Context(BaseModel):
        width: float
        height: float

    def _build(self, request: HttpRequest) -> Context: ...

    def render(self, request: HttpRequest) -> str:
        return render_to_string(
            f"app/widgets/{self.template_name}.html",
            context=self._build(request).model_dump(),
            request=request,
        )

    def update_dimensions(self, width_in_px: float, height_in_px: float) -> None:
        self.width_in_px = width_in_px
        self.height_in_px = height_in_px



from django.http import HttpRequest, HttpResponse
from django.shortcuts import render, redirect
from django.contrib import messages
from django.views.decorators.http import require_POST, require_GET

from django.contrib.auth import authenticate, login, logout


@require_GET
def login_page(request: HttpRequest) -> HttpResponse:
    return render(request, "accounts/login.html")


@require_POST
def auth_request(request: HttpRequest) -> HttpResponse:
    username = request.POST["username"]
    password = request.POST["password"]

    user = authenticate(request, username=username, password=password)

    if user is not None:
        login(request, user)
        return redirect("home")
    else:
        messages.error(request, "Login insuccessful, please try again.")
        return redirect("login-page")


def logout_view(request: HttpRequest) -> HttpResponse:
    logout(request)
    return redirect("home")



from django.http import HttpRequest
from typing import (
    Type,
    Self,
)
from enum import StrEnum

from pydantic import BaseModel, model_validator

from .base import Control
from newsflash.callback import Callback, construct_callback


class SelectContext(BaseModel):
    id: str
    options: list[str]
    selected: str
    callback: Callback | None
    swap_oob: bool

    @model_validator(mode="after")
    def check_allowed_value(self) -> Self:
        if self.selected not in self.options:
            raise ValueError("value not allowed")
        return self


class Select(Control):
    template_name: str = "control/select"
    selected: str

    def _get_options(self) -> list[str]: ...

    def _build(self, request: HttpRequest) -> SelectContext:
        assert self.id is not None

        if "on_input" in self.__class__.__dict__:
            callback = construct_callback(self.__class__.on_input, "select", "input")
        else:
            callback = None

        return SelectContext(
            id=self.id,
            options=self._get_options(),
            selected=self.selected,
            callback=callback,
            swap_oob=self.swap_oob,
        )

    def on_input(self, *args, **kwargs) -> None:
        ...


class EnumSelect(Select):
    template_name: str = "control/select"
    options: Type[StrEnum]

    def _get_options(self) -> list[str]:
        return [str(o) for o in self.options]


class ListSelect(Select):
    template_name: str = "control/select"
    options: list[str]

    def _get_options(self) -> list[str]:
        return self.options



from .base import Layout

from pydantic import BaseModel
from django.http import HttpRequest


class FlexContext(BaseModel):
    children: list[str]
    direction: str
    main_element: bool


class Flex(Layout):
    template_name: str = "layout/flex"
    direction: str

    def _build(self, request: HttpRequest) -> FlexContext:
        return FlexContext(
            children=[child.render(request) for child in self.children],
            direction=self.direction,
            main_element=self.main_element,
        )


class FlexColumns(Flex):
    direction: str = "row"


class FlexRows(Flex):
    direction: str = "col"



from django.http import HttpRequest

from pydantic import BaseModel
from .base import Text
from typing import Literal


type TextAlign = Literal["left"] | Literal["center"] | Literal["right"]


class TitleContext(BaseModel):
    title: str
    text_size: str
    align: TextAlign


class Title(Text):
    template_name: str = "text/title"
    title: str
    text_size: str = "3xl"
    align: TextAlign = "left"
    

    def update_title(self, new_title: str) -> None:
        self.title = new_title

    def _build(self, request: HttpRequest) -> TitleContext:
        return TitleContext(
            title=self.title, text_size=self.text_size, align=self.align
        )


class SubTitle(Title):
    text_size: str = "2xl"



from django.urls import path
from . import views


urlpatterns = [
    path("login", views.login_page, name="login-page"),
    path("logout", views.logout_view, name="logout-view"),
    path("auth", views.auth_request, name="auth-request"),
]



from .base import Chart, ChartContext
from .utils import chart_space_to_xy, Point, calculate_ticks
from .axes import XAxis, YAxis, format_float, order_of_magnitude
from pydantic import BaseModel
from math import ceil, floor

from django.http import HttpRequest


class SvgPoint(BaseModel):
    x: float
    y: float


class LineChartContext(ChartContext):
    points: list[SvgPoint]
    values: list[float]
    line_color: str


class LineChart(Chart):
    chart_template_name: str = "chart/line"
    x_points: list[float] = []
    y_points: list[float] = []
    svg_points: list[SvgPoint] = []

    @property
    def x_ticks(self) -> tuple[float, float]:
        major_x_tick, minor_x_tick = calculate_ticks(
            self.x_points[-1], self.x_points[0]
        )
        return major_x_tick, minor_x_tick

    @property
    def y_ticks(self) -> tuple[float, float]:
        major_y_tick, minor_y_tick = calculate_ticks(max(self.y_points), 0)
        return major_y_tick, minor_y_tick

    @property
    def max_width(self) -> float:
        max_width = self.x_points[-1] - self.x_points[0]
        return max_width

    @property
    def max_height(self) -> float:
        max_height = max(self.y_points)
        max_height = ceil(max_height / self.y_ticks[0]) * self.y_ticks[0]
        return max_height

    @property
    def y_order_of_magnitude(self) -> int:
        return order_of_magnitude(self.y_ticks[0])

    def _build_x_axis(self) -> XAxis:
        major_x_tick, minor_x_tick = self.x_ticks

        return XAxis(
            max_width=self.max_width,
            max_height=self.max_height,
            x_points=self.x_points,
            width_in_px=self.width_in_px,
            height_in_px=self.height_in_px,
            padding=self.padding,
            major_grid_lines=self.x_major_grid_lines,
            minor_grid_lines=self.x_minor_grid_lines,
            major_tick=major_x_tick,
            minor_tick=minor_x_tick,
            id=self.id + "_x_axis" if self.id is not None else None,
            order_of_magnitude=1,
            align="on_point",
            draw_major_ticks=True,
        )

    def _build_y_axis(self) -> YAxis:
        major_y_tick, minor_y_tick = self.y_ticks

        return YAxis(
            max_width=self.max_width,
            max_height=self.max_height,
            width_in_px=self.width_in_px,
            height_in_px=self.height_in_px,
            padding=self.padding,
            major_grid_lines=self.y_major_grid_lines,
            minor_grid_lines=self.y_minor_grid_lines,
            major_tick=major_y_tick,
            minor_tick=minor_y_tick,
            id=self.id + "_y_axix" if self.id is not None else None,
            order_of_magnitude=self.y_order_of_magnitude,
        )

    def set_points(
        self,
        x: list[float],
        points: list[float],
    ) -> None:
        assert len(x) == len(points)
        assert self.padding is not None

        self.x_points = x
        self.y_points = points

        self.svg_points = []
        for idx, bar in enumerate(points):
            point = chart_space_to_xy(
                self.max_width,
                self.max_height,
                self.padding,
                self.width_in_px,
                self.height_in_px,
                Point(x=idx, y=bar),
            )

            self.svg_points.append(
                SvgPoint(
                    x=point.x,
                    y=point.y,
                )
            )

        major_ticks_y = [
            (t + 1) * self.y_ticks[0]
            for t in range(floor(self.max_height / self.y_ticks[0]))
        ]
        major_ticks_y = [
            format_float(t / pow(10, self.y_order_of_magnitude)) for t in major_ticks_y
        ]
        max_len = max([len(t) for t in major_ticks_y])
        self.padding.ps = 22 + 8.5 * (max_len)

    def _build_chart(self, request: HttpRequest, id: str) -> LineChartContext:
        x_axis = self._build_x_axis()
        y_axis = self._build_y_axis()

        x_axis = x_axis.render(request)
        y_axis = y_axis.render(request)

        return LineChartContext(
            id=id,
            width=self.width_in_px,
            height=self.height_in_px,
            width_half=self.width_in_px / 2,
            points=self.svg_points,
            values=self.y_points,
            x_axis=x_axis,
            y_axis=y_axis,
            line_color="#967a3d",
            title=self.title,
            swap_oob=self.swap_oob,
        )

    def render_chart(self, request: HttpRequest, authenticated: bool) -> str:
        self.set_points(
            x=self.x_points,
            points=self.y_points,
        )
        return super().render_chart(request, authenticated)



from django.http import HttpRequest
from django.template.loader import render_to_string

from .type import Chart as ChartWidget
from .utils import Padding
from newsflash.callback import Callback, construct_callback

from pydantic import BaseModel


class ChartWrapperContext(BaseModel):
    id: str
    title: str | None
    description: str | None
    endpoint_name: str
    callback: Callback | None


class ChartContext(BaseModel):
    id: str
    width: float
    height: float
    width_half: float
    x_axis: str
    y_axis: str
    title: str | None
    swap_oob: bool


class Chart(ChartWidget):
    template_name: str = "chart/container"
    chart_template_name: str = ""
    title: str | None = None
    description: str | None = None
    endpoint_name: str = "chart"
    padding: Padding = Padding(ps=50, pt=60, pe=20, pb=30)
    x_major_grid_lines: bool = False
    x_minor_grid_lines: bool = False
    y_major_grid_lines: bool = False
    y_minor_grid_lines: bool = False
    swap_oob: bool = False
    require_login: bool = False

    def on_load(self) -> None:
        pass

    def _build(self, request: HttpRequest) -> ChartWrapperContext:
        assert self.id is not None, "A chart must have an ID."
        assert self.endpoint_name is not None

        if "on_load" in self.__class__.__dict__:
            callback = construct_callback(
                self.__class__.on_load,
                "chart",
                "load delay:50ms, resize from:window delay:250ms",
            )
        else:
            callback = None

        return ChartWrapperContext(
            id=self.id,
            title=self.title,
            description=self.description,
            endpoint_name=self.endpoint_name,
            callback=callback,
        )

    def _build_chart(self, request: HttpRequest, id: str) -> ChartContext:
        ...

    def render_chart(self, request: HttpRequest, authenticated: bool) -> str:
        if self.require_login and not authenticated:
            return self.render_unauthorized(request)

        assert self.id is not None
        context = self._build_chart(request, id=self.id)
        return render_to_string(
            f"app/widgets/{self.chart_template_name}.html",
            context=context.model_dump(),
            request=request,
        )

    def render_unauthorized(self, request: HttpRequest) -> str:
        assert self.id is not None
        return render_to_string(
            "app/unauthorized.html",
            context={"id": self.id},
            request=request,
        )



from django.http import HttpRequest

from .base import Text
from pydantic import BaseModel


class ParagraphContext(BaseModel):
    text: str


class Paragraph(Text):
    template_name: str = "text/paragraph"
    text: str

    def _build(self, request: HttpRequest) -> ParagraphContext:
        return ParagraphContext(
            text=self.text,
        )



from inspect import signature, getsourcefile
from typing import get_args, ForwardRef, Callable, Type, Literal
import json
from enum import StrEnum
from runpy import run_path

from django.http import HttpRequest, HttpResponse
from django.shortcuts import render

from newsflash import App
from newsflash.widgets.base import Widget
from newsflash.widgets.chart.base import Chart
from newsflash.widgets.control.select import Select
from newsflash.widgets.text.base import Text
from newsflash.widgets import Button


type ChartDimension = dict[Literal["width"] | Literal["height"], float]
type ChartDimensions = dict[str, ChartDimension]


def parse_chart_dimensions(request: HttpRequest) -> ChartDimensions:
    dimensions = request.POST["dimensions"]
    assert isinstance(dimensions, str)
    chart_dimensions = [
        {k.removesuffix("-container"): v for k, v in chart.items()}
        for chart in json.loads(dimensions)
    ]

    chart_dimensions_dict: ChartDimensions = {}
    for chart in chart_dimensions:
        chart_dimensions_dict.update(chart)
    
    return chart_dimensions_dict


def parse_additional_inputs(request: HttpRequest, trigger: str) -> dict[str, str]:
    additional_inputs: dict[str, str] = {}

    for k in request.POST:
        if k.endswith("-value") and k.removesuffix("-value") != trigger:
            _value = request.POST[k]
            assert isinstance(_value, str)
            additional_inputs[k.removesuffix("-value")] = _value
    
    return additional_inputs


def parse_request_inputs(request: HttpRequest) -> tuple[str, dict[str, str], ChartDimensions]:
    trigger: str = request.headers["Hx-Trigger"]  # type: ignore
    additional_inputs = parse_additional_inputs(request, trigger)
    chart_dimensions_dict = parse_chart_dimensions(request)
    return trigger, additional_inputs, chart_dimensions_dict


def resolve_forward_ref(widget_type: Type[Widget], callback_fn: Callable) -> Type[Widget]:
    if isinstance(widget_type, ForwardRef):
        source_file = getsourcefile(callback_fn)
        assert source_file is not None
        file_globals = run_path(source_file)
        widget_type = file_globals[widget_type.__forward_arg__]

    return widget_type


class WidgetIO(StrEnum):
    INPUT = "input"
    OUTPUT = "output"
    BOTH = "both"


def process_callback_arg(callback_fn: Callable, parameter: str) -> tuple[Type[Widget], str, WidgetIO]:
    sig = signature(callback_fn)
    annotation = sig.parameters[parameter].annotation
    args = get_args(annotation)

    widget_type = args[0]
    widget_type = resolve_forward_ref(widget_type, callback_fn)
    widget_id = args[1]

    try:
        widget_io = args[2]
    except IndexError:
        if issubclass(widget_type, Select) or issubclass(widget_type, Button):
            widget_io = WidgetIO.INPUT
        elif issubclass(widget_type, Chart) or issubclass(widget_type, Text):
            widget_io = WidgetIO.OUTPUT
        else:
            raise ValueError(f"unexpected widget_type: {widget_type}")
        
    return widget_type, widget_id, widget_io


def render_callback_outputs(callback_outputs: dict[str, Widget], request: HttpRequest) -> bytes:
    result: str = ""

    for callback_output in callback_outputs.values():
        if isinstance(callback_output, Chart):
            result += (" " + callback_output.render_chart(request, authenticated=request.user.is_authenticated))
        else:
            result += (" " + callback_output.render(request))
    
    return result.encode()
        

def get_callback_in_and_outputs(
    callback_fn: Callable,
    additional_inputs: dict[str, str],
    chart_dimensions_dict: ChartDimensions,
) -> tuple[dict[str, Widget], dict[str, Widget]]:
    sig = signature(callback_fn)

    callback_inputs: dict[str, Widget] = {}
    callback_outputs: dict[str, Widget] = {}
    for param in sig.parameters:
        if param == "self" or param == "selected":
            continue

        widget_type, widget_id, widget_io = process_callback_arg(callback_fn, param)
        widget_type = resolve_forward_ref(widget_type, callback_fn)

        if issubclass(widget_type, Select):
            widget = widget_type(selected=additional_inputs[widget_id])

        if issubclass(widget_type, Chart):
            widget = widget_type(
                width_in_px=chart_dimensions_dict[widget_id]["width"],
                height_in_px=chart_dimensions_dict[widget_id]["height"],
                swap_oob=True,
            )

        if issubclass(widget_type, Text):
            widget = widget_type()
            
        callback_inputs[param] = widget            
        if widget_io == WidgetIO.OUTPUT or widget_io == WidgetIO.BOTH:
            widget.swap_oob = True
            callback_outputs[param] = widget

    return callback_inputs, callback_outputs


def build_main_view(app: Type[App]) -> Callable:
    def main(request: HttpRequest, page_num: int = 1) -> HttpResponse:
        _app = app()
        return render(
            request, "app/page.html", context={"content": _app.render(request), "page_num": page_num}
        )

    return main


def build_button_view(app: Type[App]) -> Callable:

    def click(request: HttpRequest) -> HttpResponse:
        trigger, additional_inputs, chart_dimensions_dict = parse_request_inputs(request)

        _app = app()
        button_element = _app.query_one(trigger, Button)
        assert button_element is not None

        callback_inputs, callback_outputs = get_callback_in_and_outputs(
            button_element.on_click, additional_inputs, chart_dimensions_dict
        )

        button_element.on_click(**callback_inputs)
        return HttpResponse(render_callback_outputs(callback_outputs, request))

    return click


def build_select_view(app: Type[App]) -> Callable:

    def select(request: HttpRequest) -> HttpResponse:
        trigger, additional_inputs, chart_dimensions_dict = parse_request_inputs(request)
        trigger = trigger.removesuffix("-input")
        value = request.POST[f"{trigger}-value"]

        _app = app()
        select_element = _app.query_one(trigger, Select)
        assert select_element is not None

        callback_inputs, callback_outputs = get_callback_in_and_outputs(
            select_element.on_input, additional_inputs, chart_dimensions_dict
        )

        value_type = type(select_element.selected)
        select_element.selected = value_type(value)
        select_element.on_input(**callback_inputs)

        return HttpResponse(render_callback_outputs(callback_outputs, request))

    return select


def build_chart_view(app: Type[App]) -> Callable:

    def chart(request: HttpRequest) -> HttpResponse:
        trigger, additional_inputs, chart_dimensions_dict = parse_request_inputs(request)
        trigger = trigger.removesuffix("-wrapper")

        _app = app()
        chart_element = _app.query_one(trigger, Chart)
        assert chart_element is not None

        chart_element.width_in_px = chart_dimensions_dict[trigger]["width"]
        chart_element.height_in_px = chart_dimensions_dict[trigger]["height"]

        callback_inputs, callback_outputs = get_callback_in_and_outputs(
            chart_element.on_load, additional_inputs, chart_dimensions_dict
        )

        callback_outputs[trigger] = chart_element
        chart_element.on_load(**callback_inputs)

        return HttpResponse(render_callback_outputs(callback_outputs, request))

    return chart



from django.http import HttpRequest

from pydantic import BaseModel
from .base import Text


class Notification(BaseModel):
    text: str
    duration_in_ms: int


class NotificationsContext(BaseModel):
    notifications: list[Notification]


class Notifications(Text):
    template_name: str = "text/notification"
    notifications: list[Notification] = []

    def push(self, text: str, duration_in_ms: int = 2000) -> None:
        self.notifications.append(
            Notification(
                text=text,
                duration_in_ms=duration_in_ms,
            )
        )

    def _build(self, request: HttpRequest) -> NotificationsContext:
        return NotificationsContext(
            notifications=self.notifications
        )



from .base import Chart, ChartContext
from .utils import Point, chart_space_to_xy, calculate_ticks
from .axes import XAxis, YAxis, format_float, order_of_magnitude
from pydantic import BaseModel
from math import floor, ceil

from django.http import HttpRequest


class SvgRect(BaseModel):
    x: float
    y: float
    width: float
    height: float
    color: str


class BarChartContext(ChartContext):
    bars: list[SvgRect]
    values: list[float]


class BarChart(Chart):
    chart_template_name: str = "chart/bar"
    labels: list[str] = []
    bars: list[float] = []
    major_x_tick: int | None = 5
    minor_x_tick: int | None = None
    major_y_tick: int | None = 10
    minor_y_tick: int | None = None

    svg_rects: list[SvgRect] = []
    x_axis: XAxis | None = None
    y_axis: YAxis | None = None

    @staticmethod
    def calculate_bar_width(
        width_in_px: float, max_width: float, ps: float, pe: float, frac: float
    ) -> float:
        chart_width_px = width_in_px - ps - pe
        width_per_bar = chart_width_px / max_width
        return width_per_bar * frac

    @staticmethod
    def calculate_bar_height(
        height_in_px: float, max_height: float, pt: float, pb: float, bar: float
    ) -> float:
        chart_height_px = height_in_px - pt - pb
        bar_height = (bar / max_height) * chart_height_px
        return bar_height

    @property
    def y_ticks(self) -> tuple[float, float]:
        major_y_tick, minor_y_tick = calculate_ticks(max(self.bars), 0)
        return major_y_tick, minor_y_tick

    @property
    def max_height(self) -> float:
        max_height = max(self.bars)
        max_height = ceil(max_height / self.y_ticks[0]) * self.y_ticks[0]
        return max_height

    @property
    def y_order_of_magnitude(self) -> int:
        return order_of_magnitude(self.y_ticks[0])

    def _build_x_axis(self) -> XAxis:
        major_x_tick, minor_x_tick = calculate_ticks(len(self.bars), 0)

        return XAxis(
            max_width=len(self.bars),
            max_height=self.max_height,
            x_points=list(range(len(self.bars))),
            x_labels=self.labels,
            width_in_px=self.width_in_px,
            height_in_px=self.height_in_px,
            padding=self.padding,
            major_grid_lines=self.x_major_grid_lines,
            minor_grid_lines=self.x_minor_grid_lines,
            major_tick=major_x_tick,
            minor_tick=minor_x_tick,
            id=self.id + "_x_axis" if self.id is not None else None,
            order_of_magnitude=1,
            align="between",
            draw_major_ticks=False,
        )

    def _build_y_axis(self) -> YAxis:
        major_y_tick, minor_y_tick = calculate_ticks(max(self.bars), 0)
        return YAxis(
            max_width=len(self.bars),
            max_height=self.max_height,
            width_in_px=self.width_in_px,
            height_in_px=self.height_in_px,
            padding=self.padding,
            major_grid_lines=self.y_major_grid_lines,
            minor_grid_lines=self.y_minor_grid_lines,
            major_tick=major_y_tick,
            minor_tick=minor_y_tick,
            id=self.id + "_y_axis" if self.id is not None else None,
            order_of_magnitude=self.y_order_of_magnitude,
        )

    def set_bars(
        self,
        labels: list[str],
        bars: list[float],
    ) -> None:
        assert self.padding is not None

        self.labels = labels
        self.bars = bars

        bar_width_frac = 0.85
        bar_width = self.calculate_bar_width(
            self.width_in_px,
            len(bars),
            self.padding.ps,
            self.padding.pe,
            bar_width_frac,
        )

        self.svg_rects = []
        for idx, bar in enumerate(bars):
            point = chart_space_to_xy(
                len(bars),
                self.max_height,
                self.padding,
                self.width_in_px,
                self.height_in_px,
                Point(x=idx, y=bar),
            )

            point.x = point.x + (1 - bar_width_frac) * bar_width / 2

            bar_height = self.calculate_bar_height(
                self.height_in_px,
                self.max_height,
                self.padding.pt,
                self.padding.pb,
                bar,
            )

            self.svg_rects.append(
                SvgRect(
                    x=point.x,
                    y=point.y,
                    width=bar_width,
                    height=bar_height,
                    color="#8839ef",
                )
            )

        major_ticks_y = [
            (t + 1) * self.y_ticks[0]
            for t in range(floor(self.max_height / self.y_ticks[0]))
        ]
        major_ticks_y = [
            format_float(t / pow(10, self.y_order_of_magnitude)) for t in major_ticks_y
        ]
        max_len = max([len(t) for t in major_ticks_y])
        self.padding.ps = 22 + 8.5 * (max_len)

    def _build_chart(
        self,
        request: HttpRequest,
        id: str,
    ) -> BarChartContext:
        x_axis = self._build_x_axis()
        y_axis = self._build_y_axis()

        x_axis = x_axis.render(request)
        y_axis = y_axis.render(request)

        return BarChartContext(
            id=id,
            width=self.width_in_px,
            height=self.height_in_px,
            width_half=self.width_in_px / 2,
            bars=self.svg_rects,
            values=self.bars,
            x_axis=x_axis,
            y_axis=y_axis,
            title=self.title,
            swap_oob=self.swap_oob,
        )

    def render_chart(self, request: HttpRequest, authenticated: bool) -> str:
        self.set_bars(
            labels=self.labels,
            bars=self.bars,
        )
        return super().render_chart(request, authenticated)



from ..base import Widget
from typing import Type, TypeVar

from pydantic import BaseModel
from django.http import HttpRequest


T = TypeVar("T")

class LayoutContext(BaseModel):
    children_and_span: list[tuple[str, int]]
    grow: bool
    main_element: bool


class Layout(Widget):
    main_element: bool = False
    children: list[Widget]
    grow: bool = False

    def get_all_children(self) -> list[Widget]:
        graphs: list[Widget] = []

        for child in self.children:
            if isinstance(child, Layout):
                graphs.append(child)
                graphs.extend(child.get_all_children())
            else:
                graphs.append(child)

        return graphs

    def query_one(self, id: str, type: Type[T]) -> T | None:
        for child in self.children:
            if child.id == id and isinstance(child, type):
                return child
            elif isinstance(child, Layout):
                child_query_one_result = child.query_one(id, type)
                if child_query_one_result is not None:
                    return child_query_one_result

        return None

    def query_type(self, id: str) -> Type[Widget] | None:
        for child in self.children:
            if child.id == id:
                return type(child)
            elif isinstance(child, Layout):
                child_query_type_result = child.query_type(id)
                if child_query_type_result is not None:
                    return child_query_type_result
                
    def _build(self, request: HttpRequest) -> LayoutContext:
        return LayoutContext(
            children_and_span=list(
                zip(
                    [c.render(request) for c in self.children],
                    [c.span for c in self.children],
                )
            ),
            grow=self.grow,
            main_element=self.main_element,
        )



from .text import Title, Paragraph, Notifications
from .layout import Rows, Columns, FlexRows, FlexColumns
from .control import Button, EnumSelect, ListSelect
# from .chart import BarChart, LineChart


__all__ = [
    # Text
    "Title",
    "Paragraph",
    "Notifications",
    
    # Layout
    "Rows",
    "Columns",
    "FlexRows",
    "FlexColumns",

    # Control
    "Button",
    "EnumSelect",
    "ListSelect",

    # Charts
    # "BarChart",
    # "LineChart",
]


from typing import Type
from django.urls import path
from . import views
from newsflash import App


urlpatterns = []


def set_urlpatterns(app: Type[App]) -> None:
    global urlpatterns

    urlpatterns = [
        path("", views.build_main_view(app), name="home"),
        path("<int:page_num>", views.build_main_view(app), name="home"),
        path("click", views.build_button_view(app), name="click"),
        path("select", views.build_select_view(app), name="select"),
        path("chart", views.build_chart_view(app), name="chart"),
    ]



from typing import Literal

from django.http import HttpRequest
from pydantic import BaseModel

from math import floor, log10
from ..base import Widget
from .base import Padding
from .utils import Point, chart_space_to_xy


def format_float(x: float) -> str:
    return str(int(x)) if x.is_integer() else str(x)


def order_of_magnitude(x: float):
    if x == 0:
        return 0
    return floor(log10(abs(x)))


class AxisContext(BaseModel):
    graph_width_px: float
    graph_height_px: float
    padding: Padding
    major_ticks: list[tuple[Point, str]]
    minor_ticks: list[tuple[Point, str]] | None
    draw_major_ticks: bool = True
    major_grid_lines: bool
    minor_grid_lines: bool


class Axis(Widget):
    max_width: float
    max_height: float
    padding: Padding
    major_tick: float
    minor_tick: float | None
    major_grid_lines: bool
    minor_grid_lines: bool
    draw_major_ticks: bool = True
    order_of_magnitude: int


class YAxisContext(AxisContext):
    multiplier: int


class YAxis(Axis):
    template_name: str = "chart/y_axis"

    def _build_ticks(self, tick: float) -> list[tuple[Point, str]]:
        major_ticks_y = [t * tick for t in range(floor(self.max_height / tick) + 1)]
        major_ticks_points = [
            chart_space_to_xy(
                self.max_width,
                self.max_height,
                self.padding,
                self.width_in_px,
                self.height_in_px,
                Point(x=0, y=y),
            )
            for y in major_ticks_y
        ]
        major_ticks_y = [
            format_float(t / pow(10, self.order_of_magnitude)) for t in major_ticks_y
        ]
        return list(zip(major_ticks_points, major_ticks_y))

    def _build(self, request: HttpRequest) -> YAxisContext:
        if self.minor_tick is not None:
            minor_ticks = self._build_ticks(self.minor_tick)
        else:
            minor_ticks = None
        minor_ticks = None

        return YAxisContext(
            graph_width_px=self.width_in_px - self.padding.ps - self.padding.pe,
            graph_height_px=self.height_in_px - self.padding.pt - self.padding.pb,
            padding=self.padding,
            major_ticks=self._build_ticks(self.major_tick),
            minor_ticks=minor_ticks,
            major_grid_lines=self.major_grid_lines,
            minor_grid_lines=self.minor_grid_lines,
            multiplier=pow(10, self.order_of_magnitude),
        )


class XAxis(Axis):
    template_name: str = "chart/x_axis"

    x_points: list[float]
    x_labels: list[str] | None = None
    align: Literal["on_point"] | Literal["between"]

    def _build_ticks(self, tick: float) -> list[tuple[Point, str]]:
        major_ticks_x = [x for x in self.x_points if (x % tick == 0)]
        major_ticks_points = [
            chart_space_to_xy(
                self.max_width,
                self.max_height,
                self.padding,
                self.width_in_px,
                self.height_in_px,
                Point(x=x - self.x_points[0], y=0),
            )
            for x in major_ticks_x
        ]
        if self.x_labels:
            major_ticks_x = self.x_labels
        else:
            major_ticks_x = [format_float(t) for t in major_ticks_x]

        if self.align == "between":
            shift = (major_ticks_points[1].x - major_ticks_points[0].x) / 2
            major_ticks_points = [
                Point(x=p.x + shift, y=p.y) for p in major_ticks_points
            ]

        return list(zip(major_ticks_points, major_ticks_x))

    def _build(self, request: HttpRequest) -> AxisContext:
        self.minor_tick = None
        if self.minor_tick is not None:
            minor_ticks = self._build_ticks(self.minor_tick)
        else:
            minor_ticks = None

        return AxisContext(
            graph_width_px=self.width_in_px - self.padding.ps - self.padding.pe,
            graph_height_px=self.height_in_px - self.padding.pt - self.padding.pb,
            padding=self.padding,
            major_ticks=self._build_ticks(self.major_tick),
            minor_ticks=minor_ticks,
            draw_major_ticks=self.draw_major_ticks,
            major_grid_lines=self.major_grid_lines,
            minor_grid_lines=self.minor_grid_lines,
        )
